use std::iter::once;

use crate::common::SectionInfo;
use crate::common::{Relocatable, Section};
use crate::error::Error;

pub enum SegmentType {
    Text,
    Data,
    Bss,
}

type SegmentSections = Vec<SectionInfo>;

#[derive(Default)]
pub struct Loadable {
    text: SegmentSections,
    rodata: SegmentSections,
    data: SegmentSections,
    bss: SegmentSections,
}

impl Loadable {
    pub fn process_section(&mut self, section: Section) -> Result<(), Error> {
        match section {
            Section::Text(si) => self.text.push(si),
            Section::Data(si, true) => self.rodata.push(si),
            Section::Data(si, false) => self.data.push(si),
            Section::Bss(si) => self.bss.push(si),
        }

        Ok(())
    }

    pub fn sections_mut(&mut self) -> impl Iterator<Item = (&'static str, &mut SegmentSections)> {
        let text = once(&mut self.text).map(|s| (".text", s));
        let rodata = once(&mut self.rodata).map(|s| (".rodata", s));
        let data = once(&mut self.data).map(|s| (".data", s));
        let bss = once(&mut self.bss).map(|s| (".bss", s));

        text.chain(rodata).chain(data).chain(bss)
    }

    pub fn program_bytes<'l, R: Relocatable>(
        &'l self,
        objects: &'l [R],
    ) -> impl Iterator<Item = Result<&'l [u8], Error>> {
        let text_iter = self.text.iter();
        let rodata_iter = self.rodata.iter();

        let data1 = text_iter.filter_map(move |si| match si.data_index {
            None => None,
            Some(di) => Some(objects[di.object_index].section_data(di)),
        });
        let data2 = rodata_iter.filter_map(move |si| match si.data_index {
            None => None,
            Some(di) => Some(objects[di.object_index].section_data(di)),
        });

        data1.chain(data2)
    }

    pub fn data_bytes<'l, R: Relocatable>(
        &'l self,
        objects: &'l [R],
    ) -> impl Iterator<Item = Result<&'l [u8], Error>> {
        let iter = self.data.iter().filter_map(move |si| match si.data_index {
            None => None,
            Some(di) => Some(objects[di.object_index].section_data(di)),
        });

        iter
    }

    // .text + .rodata
    pub fn program_size(&self) -> u64 {
        let text_iter = self.text.iter();
        let rodata_iter = self.rodata.iter();

        let mut result = 0u64;
        for section in text_iter.chain(rodata_iter) {
            result += section.data_size
        }

        result
    }

    // data
    pub fn data_size(&self) -> u64 {
        let mut result = 0u64;
        for section in self.data.iter() {
            result += section.data_size
        }

        result
    }

    // bss
    pub fn bss_size(&self) -> u64 {
        let mut result = 0u64;
        for section in self.bss.iter() {
            result += section.data_size
        }

        result
    }
}
