use crate::error::Error;
use std::{
    convert::TryFrom,
    fmt::{Display, Formatter},
};

use super::DataIndex;

#[derive(Clone)]
pub struct SectionInfo {
    pub file_size: u64,
    pub data_size: u64,
    pub data_index: Option<DataIndex>, // some form of indexing to the source data (Relocatable's data)
    pub offset: u64,
}

#[derive(Clone)]
pub enum Section {
    Text(SectionInfo),
    Data(SectionInfo, bool), // readonly bool
    Bss(SectionInfo),
}

impl<'data> TryFrom<(&str, SectionInfo)> for Section {
    type Error = Error;

    fn try_from(value: (&str, SectionInfo)) -> Result<Self, Self::Error> {
        if value.0.starts_with(".text") {
            Ok(Section::Text(value.1))
        } else if value.0.starts_with(".rodata") {
            Ok(Section::Data(value.1, true))
        } else if value.0.starts_with(".data") {
            Ok(Section::Data(value.1, false))
        } else if value.0.starts_with(".bss") {
            Ok(Section::Bss(value.1))
        } else {
            Err(Error::InvalidSectionName)
        }
    }
}

impl Display for SectionInfo {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "file_size: {}, data_size: {}, data: {}, offset: {}",
            self.file_size,
            self.data_size,
            self.data_index.is_some(),
            self.offset,
        )
    }
}

impl Display for Section {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            Section::Text(s) => writeln!(f, "TEXT {}", s),
            Section::Data(s, false) => writeln!(f, "DATA {}", s),
            Section::Data(s, true) => writeln!(f, "RO-DATA {}", s),
            Section::Bss(s) => writeln!(f, "BSS {}", s),
        }
    }
}
